package CTL;

import CTL.Comm.*;
import CTL.Streams.*;
import CTL.Types.*;

import java.io.*;
import java.lang.reflect.*;
import java.nio.*;

/** Incoming connection handling thread */
public class ConnectHandler extends Thread
{
	private Env env = null;

	public ConnectHandler (Env env)
	{
		this.env = env;
	}

	/** Handles a connection to this process
	 */
	public void run ()
	{
		Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
		
		// TODO: Determine ID somehow
		int conn = (env.grp.myID() == 0 ? 1 : 0);
		
		try
		{
			accept(conn);
		}
		catch (Exception e)
		{
			if (!env.grp.running())
				return;
			//e.printStackTrace();
			RUtil.except(e);
			try
			{
				env.comm[conn].close();
			} catch (IOException e2) {}
		}
	}
	
	/** Accept a connection and delegate it to a handling method
	 * @param id Logical ID of the connecting client
	 */
	public void accept (int id) throws IOException, IllegalAccessException,
		InvocationTargetException, CTLException, NoSuchMethodException,
		ClassNotFoundException, InstantiationException
	{
		if (CTL.Globals.timingHack)
			CTL.Timer.sleep(1000);

		byte[] msg = env.comm[id].recv();
		
		if (msg == null)
		{
			if (!env.grp.running())
				return;
			if (env.daemon)
				return;
			env.log.msg(Logger.ERR, "No data received.");
		}
			
		BuffyIn in = new BuffyIn(msg);
		Header ohead = null;
		long objID;
		
		ohead = in.i.serialRead(Header.class);
		env.log.msg(Logger.INFO, "Recv header: "+ohead);
		if (ohead.needSwap())
		{
			//env.log.msg(Logger.INFO, "swapping");
			in.changeEndianess(ByteOrder.LITTLE_ENDIAN);
		}

		switch (ohead.tag())
		{
			case Remote.OPER:
				env.log.msg(Logger.ERR, "OPER is not yet supported.");
				System.exit(5);
				break;

			case Remote.DAT:
				env.log.msg(Logger.ERR, "DAT is not yet supported.");
				System.exit(5);
				break;
			
			case Remote.RMI:
				objID = in.i.readLong();
				procRMI(in.i, ohead, objID, id, env);
				break;

			case Remote.CTRL:
				env.log.msg(Logger.ERR, "CTRL is not yet supported.");
				System.exit(5);
				break;

			default:
				objID = in.i.readLong();
		
				env.log.msg(Logger.INFO, ".:. Received RMI answer .:.");
				env.log.msg(Logger.INFO, "\tHeader: "+ohead);
				env.log.msg(Logger.INFO, "\tObjectID: "+objID);

				Object obj = env.map.getObj((int)objID);
				if (obj==null)
				{
					env.log.msg(Logger.ERR, "Object "+objID+" not found.");
					FID fid = in.i.serialRead(FID.class);
					// TODO: Does not work
					CTL_Throw(in.i, 1, ohead, fid, "Object \""+objID+"\" not found.",
						env);
					return;
				}

				if (!obj.getClass().getName().equals("CTL.rResult"))
				{
					procRMI(in.i, ohead, objID, 1, env);
					return;
				}

				Except ex = in.i.serialRead(Except.class);
				if (ex.name().length()>0)
					throw new CTLException("Class not found!");

				rResult res = (rResult)obj;	
				for (int i=0;i<res.results();i++)
					res.recv(in.i, i);
				env.log.msg(Logger.DBG, "Received rResult "+objID+": "+res);
				break;
		}
	}

	/** Calculates how long the argument list of a call is
	 * @param head Header of this package
	 * @param fid FunctionID of the invoked method
	 * @return Length of the argument list in bytes
	 */
	public static long argSize (Header head, FID fid)
	{
		return head.getSize()-47-fid.name().length();
	}


	/** Process an RMI call
	 * @param in Input stream for this connection
	 * @param head Header of this package
	 * @param objID Requested ObjectID
	 * @param member Local group member ID
	 * @param env Environment information
	 */
	public static void procRMI (OIStream in, Header head, long objID,
		int member, Env env) throws IOException, IllegalAccessException, 
		NoSuchMethodException, InvocationTargetException, 
		ClassNotFoundException, CTLException, InstantiationException
	{
		procRMI(in, head, objID, member, env, null);
	}
	
	/** Process an RMI call
	 * @param in Input stream for this connection
	 * @param head Header of this package
	 * @param objID Requested ObjectID
	 * @param member Local group member ID
	 * @param env Environment information
	 * @param suffix Custom suffix to use for called Classes
	 */
	public static void procRMI (OIStream in, Header head, long objID, 
		int member, Env env, String suffix) throws IOException, IllegalAccessException, 
		NoSuchMethodException, InvocationTargetException, 
		ClassNotFoundException, CTLException, InstantiationException
	{
		FID fid = in.serialRead(FID.class);
		Class targetClass = null;

		String className = fid._class();
		if (suffix != null)
			className = fid._class(suffix);

		try
		{
			env.log.msg(Logger.INFO, "FID: ObjectID: "+objID+
				" - Target class: "+className);
			//env.log.msg(Logger.INFO, env.map.dump());
			
			if (objID<=0)
				targetClass = Class.forName(className);
			else
			{
				Object req = env.map.getObj((int)objID);
				if (req==null)
					throw new ClassNotFoundException("No such ObjectID.");
				targetClass = RUtil.classToRI(req.getClass());
			}
		}
		catch (ClassNotFoundException e)
		{
			try
			{
				targetClass = Env.TypeDefs.get(className);
			} catch (ClassNotFoundException e2) {}

			if (targetClass == null)
			{
				if ((suffix != null) && (suffix.equals("Debug")))
					throw new ClassNotFoundException("Could not determine class.");
				env.log.msg(Logger.ERR, "Class not found: \""+
					className+"\" ("+e.getMessage()+")");
				CTL_Throw(in, member, head, fid, "Class \""+className+"\" not found.",
					env);
				return;
			}
		}
		
		env.log.msg(Logger.DBG, "ObjectID: "+objID+" - Target class: "+
			targetClass.getName());

		Class[] args = new Class[6];
		args[0] = OIStream.class;
		args[1] = Header.class;
		args[2] = Long.TYPE;
		args[3] = FID.class;
		args[4] = Integer.TYPE;
		args[5] = Env.class;
		
		Object[] args2 = new Object[6];
		args2[0] = (Object)in;
		args2[1] = (Object)head;
		args2[2] = new Long(objID);
		args2[3] = (Object)fid;
		env.log.msg(Logger.DBG, ""+member);
		env.log.msg(Logger.DBG, ""+env.grp);
		args2[4] = new Integer(member);
		args2[5] = (Object)env;

		Method m = targetClass.getDeclaredMethod("accept", args);
		m.invoke(null, args2);
	}

	// TODO: document
	protected static void CTL_Throw (OIStream in, int member, Header head, FID fid, 
		String msg, Env env) throws IOException, IllegalAccessException,
		ClassNotFoundException, InvocationTargetException, InstantiationException,
		CTLException
	{
		// Read junk away
		int[] sz = new int[2];
		sz[0] = (int)argSize(head,fid);
		sz[1] = (int)(argSize(head, fid) >> 32);
		for (int i=0;i<2;i++)
		{
			if (sz[i] < 0)
				continue;
			byte[] junk = new byte[sz[i]];
			in.read(junk);
		}
			
		rPointer rp = in.serialRead(rPointer.class);
		int tag = in.readInt();
			
		Header head2 = new Header(0, tag, env.grp.myInfo());
		Except ex = new Except(msg);
		Remote.answer(env.comm[member], head2, rp.objID(), ex, null);
	}
}
