package CTL.Types;

import CTL.Serialize.*;
import java.io.*;
import java.lang.reflect.*;

/** Unoptimized trivial graph class; show case for CTL.Reference
 * 
 * Bandwidth usage: 6n + 18m + 9
 * 	n: number of nodes; m: number of edges
 * 	assumption: each node has at least one edge connected to it
 *
 * Sending data and matrix directly: n^2 + 2n + 12
 */
public class Graph implements Writable
{
	private Node nodes[] = null;
	private boolean adjM[][] = null;
	private int edges = 0;

	public Graph (int size)
	{
		if (size<=0)
			return;
		resize(size);
	}

	public int size ()
	{
		if (nodes == null)
			return 0;
		return nodes.length;
	}

	public boolean equals (Object moo)
	{
		if ((moo instanceof Graph) && toString().equals(moo.toString()))
			return true;
		return false;
	}
	
	private void resize (int size)
	{
		nodes = new Node[size];
		adjM = new boolean[size][size];

		for (int i=0;i<size;i++)
			for (int j=0;j<size;j++)
				adjM[i][j] = (i==j) ? true : false;
	}

	public boolean addNode (Node data)
	{
		if (nodes==null)
		{
			resize(1);
			setNode(0, data);
			return true;
		}

		if (findNode(data)>-1)
			return false;

		Node old_nodes[] = nodes;
		boolean old_adjM[][] = adjM;
		edges = 0;
		resize(old_nodes.length+1);
		for (int i=0;i<old_nodes.length;i++)
		{
			setNode(i, old_nodes[i]);
			for (int j=0;j<old_nodes.length;j++)
				if (old_adjM[i][j])
					addEdge(i, j);
		}
		setNode(old_nodes.length, data);
		return true;
	}

	public int findNode (Node data)
	{
		if (data!=null && nodes!=null)
		{
			for (int i=0;i<nodes.length;i++)
				if (data.equals(nodes[i]))
					return i;
		}
		return -1;
	}

	public boolean isEdge (int i, int j)
	{
		return adjM[i][j];
	}

	public int cost (int i, int j)
	{
		return nodes[i].cost() + nodes[j].cost();
	}

	public boolean addEdge (int i, int j)
	{
		if (i<0 || j<0 || adjM[i][j] || i==j || 
			i>=adjM.length || j>=adjM[i].length)
			return false;
		adjM[i][j] = true;
		edges++;
		return true;
	}

	public boolean addEdge2 (int i, int j)
	{
		return addEdge(i, j) && addEdge(j, i);
	}

	public void setNode (int i, Node data)
	{
		nodes[i] = data;
	}

	public Node node (int i)
	{
		return nodes[i];
	}

	public String toString ()
	{
		if (nodes==null) return "";
		StringBuffer buf = new StringBuffer();
		for (int i=0;i<nodes.length;i++)
		{
			if (nodes[i]==null) continue;
			buf.append(nodes[i]+" ("+nodes[i].cost()+") --> ");
			for (int j=0;j<nodes.length;j++)
				if ((i!=j) && adjM[i][j])
					buf.append(nodes[j]+"  ");
			buf.append("\n");
		}
		return buf.toString();
	}
	
	public void write (SerialOut out) throws IOException, IllegalAccessException,
		InvocationTargetException, CTLException
	{
		out.writeInt(edges);
		if (nodes == null)
			return;
		
		for (int i=0;i<nodes.length;i++)
			for (int j=0;j<nodes.length;j++)
				if (i!=j && adjM[i][j])
				{
					out.serialWrite(new Reference<Node>(nodes[i]));
					out.serialWrite(new Reference<Node>(nodes[j]));
				}
	}
	
	public void read (SerialIn in) throws IOException, ClassNotFoundException, 
		InstantiationException, IllegalAccessException, InvocationTargetException
	{
		int len = in.readInt();
		//System.out.println(len);

		for (int i=0;i<len;i++)
		{
			Reference<Node> ref = (Reference<Node>)in.serialRead(
					Reference.class, Node.class);
			Reference<Node> ref2 = (Reference<Node>)in.serialRead(
					Reference.class, Node.class);
			Node n1 = (Node)ref.obj();
			Node n2 = (Node)ref2.obj();

			//System.out.println(n1+" "+n2);

			if (ref.first()) addNode(n1);
			if (ref2.first()) addNode(n2);
			addEdge(findNode(n1), findNode(n2));
		}
	}
}
