package CodeGen;

import CTL.Annotate.*;
import ReflWrap.*;

import java.lang.reflect.*;
import java.util.*;

/** Helper class for keeping track of methods inside CodeGen */ 
public class Methode
{
	/** The declaring class */
	private ClassInfo cinfo;
	/** Reflection method object */
	private Method meth;
	/** Info object */
	private MethodInfo minfo;
	/** Whether or not the method is const (== all parameters are const) */
	private boolean _const = false;
	/** Return value */
	private ClassInfo retVal = null;

	/** Default constructor
	 * @param cinfo Declaring class
	 * @param meth Method
	 */
	public Methode (ClassInfo cinfo, Method meth)
	{
		this.cinfo = cinfo;
		this.meth = meth;
		minfo = new MethodInfo(meth);
		if (cinfo != null)
			_const = cinfo.isAnot(const_.class);
		retVal = minfo.returns();
	}

	// TODO: document
	public void setReturnValue (ClassInfo retVal)
	{
		this.retVal = retVal;
	}

	/** Whether or not the method is const
	 * @return True if const, false otherwise
	 */
	public boolean _const ()
	{
		return _const;
	}

	/** Whether or not this method is exported
	 * @return False if exported, true otherwise 
	 */
	public boolean dont ()
	{
		return !meth.isAnnotationPresent(export.class);
	}

	/** Number of parameters this method takes
	 * @return Number of parameters
	 */
	public int len ()
	{
		return minfo.params().length;
	}

	/** Whether or not this method was declared in the class itself or inherited
	 * from superclasses
	 * @return True if declared locally, false otherwise
	 */
	public boolean here ()
	{
		return minfo._class()==cinfo._class();
	}

	/** Whether or not this method is void
	 * @return True if void, false otherwise
	 */
	public boolean isVoid ()
	{
		return retVal.name().equals("void");
	}

	/** Whether or not this method is static
	 * @return True if static, false otherwise
	 */
	public boolean isStatic ()
	{
		return minfo.isStatic();
	}

	/** Generate a method header
	 * @param isrResult If true, the header for an rResult wrapper method will be
	 * generated
	 * @return Method header as String
	 */
	public String head (boolean isrResult)
	{
		if (isrResult)
			return minfo.modifiersToString()+
				(isVoid() ? "void " : "rResult ")+
				minfo.name()+"_rr ("+minfo.paramsToString();
		return minfo.toString();			
	}

	/** Name of the method
	 * @return String
	 */
	public String name ()
	{
		return minfo.name();
	}

	/** Modifiers (public, private, ...) of this method
	 * @return String
	 */
	public String modifiers ()
	{
		return minfo.modifiersToString();
	}

	/** Parameters of this method
	 * @return String
	 */
	public String params ()
	{
		return minfo.paramsToString();
	}

	/** Return type of this method
	 * @return String
	 */
	public String returns ()
	{
		return retVal.name();
	}

	/** Return type with template parameters
	 * @return String
	 */
	public String returnsT ()
	{
		return minfo.returnsT();
	}

	/** The info object
	 * @return MethodInfo
	 */
	public MethodInfo minfo ()
	{
		return minfo;
	}

	/** Generate Methode-objects for all methods of a class
	 * @param cinfo Java class
	 * @return Array of Methode-objects
	 */
	public static Methode[] fetchAll (ClassInfo cinfo)
	{
		LinkedList<Methode> list = new LinkedList<Methode>();
		
		for (Method m : cinfo.methods())
		{
			if (m == null)
				list.add(null);
			else
				list.add(new Methode(cinfo, m));
		}

		Methode[] ret = new Methode[list.size()];
		for (int i=0;i<ret.length;i++)
			ret[i] = list.get(i);
		
		return ret;
	}
}
