package ReflWrap;

import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/** This class provides helper functions to the ReflWrap package */
public class Refl
{
	/** Checks if a certain annotation is present
	 * @param comp Annotation to check for
	 * @param anot Array of annotations to check against
	 * @return True if one of the annotations matches comp, false otherwise
	 */
	public static boolean hasAnot (String comp, Annotation[] anot)
	{
		for (int i=0;i<anot.length;i++)
			if (anot[i].toString().equals(comp))
				return true;
		return false;
	}

	/** Temporary conversion hack
	 * @param klass Class to convert
	 * @return Converted code
	 */
	public static String rtype (Class klass)
	{
		StringTokenizer str = new StringTokenizer(klass.toString(), ".");
		String ret = null;
		while (str.hasMoreTokens())
			ret = str.nextToken();
		if (ret==null)
			return null;
		ret = Refl.convert(ret);
		if (ret.equals("Boolean"))
			return "boolean";
		if (ret.equals("Byte"))
			return "byte";
		if (ret.equals("Char")) // Character?
			return "char";
		if (ret.equals("Double"))
			return "double";
		if (ret.equals("Float"))
			return "float";
		if (ret.equals("Int"))
			return "int";
		if (ret.equals("Long"))	
			return "long";
		if (ret.equals("Short"))
			return "short";
		// TODO: Pretty hackish, but works
		if (ret.charAt(ret.length()-1) == ';')
			ret = "Array";
		return ret;
	}
	
	/** Returns the class part of a fully-qualified typename
	 * @param type Fully-qualified typename
	 * @return Class part of the name
	 */
	public static String type (String type)
	{
		type = lastToken(type, " ");
		if (type.toCharArray()[0]=='[')
			return handleArray(type);
		StringTokenizer tok = new StringTokenizer(type, ".");
		String retval = type;
		while (tok.hasMoreTokens())
			retval = tok.nextToken();
		return retval;
	}
	
	/** Handles arrays, which are mangled by the Reflection API
	 * @param type Type
	 * @return A cleaned-up type
	 */
	public static String handleArray (String type)
	{
		if (type.equals("[Z"))
			return "boolean[]";
		if (type.equals("[B"))
			return "byte[]";
		if (type.equals("[I"))
			return "int[]";
		if (type.equals("[J"))
			return "long[]";
		if (type.equals("[S"))
			return "short[]";
		if (type.equals("[D"))
			return "double[]";
		if (type.equals("[F"))
			return "float[]";
		if (type.equals("[C"))
			return "char[]";
		if ((type.length()>2)&&(type.substring(0, 2).equals("[L")))
			return type.substring(2, type.length()-1)+"[]";
		return type;
	}

	/** Helper function for handling arrays
	 * @param type Type
	 * @return A cleaned-up type
	 */
	public static String arrayType (String type)
	{
		type = handleArray(type);
		StringTokenizer tok = new StringTokenizer(type, "[");
		return tok.nextToken();
	}

	/** Array handling hack
	 * @param type Type
	 * @return Array type
	 */
	public static String arrayType2 (String type)
	{
		String moo = arrayType(type);
		if (moo.equals("boolean"))
			return "Boolean.TYPE";
		if (moo.equals("byte"))
			return "Byte.TYPE";
		if (moo.equals("char"))
			return "Character.TYPE";
		if (moo.equals("double"))
			return "Double.TYPE";
		if (moo.equals("float"))
			return "Float.TYPE";
		if (moo.equals("int"))
			return "Integer.TYPE";
		if (moo.equals("long"))
			return "Long.TYPE";
		if (moo.equals("short"))
			return "Short.TYPE";
		return moo+".class";
	}

	/** Get a Class object from a given string
	 * @param type Requested class
	 * @return Class object
	 */
	public static Class getClass (String type)
	{
		type = type(type);
		
		if (type.equals("int")||type.equals("Int")||type.equals("Integer"))
			return Integer.TYPE;
		if (type.equals("long")||type.equals("Long"))
			return Long.TYPE;
		if (type.equals("short")||type.equals("Short"))
			return Short.TYPE;
		if (type.equals("float")||type.equals("Float"))
			return Float.TYPE;
		if (type.equals("double")||type.equals("Double"))
			return Double.TYPE;
		if (type.equals("boolean")||type.equals("Boolean"))
			return Boolean.TYPE;
		if (type.equals("byte")||type.equals("Byte"))
			return Byte.TYPE;
		if (type.equals("char")||type.equals("Char")||type.equals("Character"))
			return Character.TYPE;
		if (type.equals("String"))
			return String.class;
		if ((type.equals("Object"))||(type.equals("Array")))
			return Object.class;

		return null;
	}

	/** Retrieve a code snippet for generating a new object
	 * @param moo Class to generate an object for
	 * @return Code
	 */
	public static String class_ (Class moo)
	{
		ClassInfo cinfo = new ClassInfo(moo);
		return cinfo.clstr();	
	}
	
	/** Get the default value for a specific type
	 * @param type Type
	 * @return Default value
	 */
	public static String defaultVal (String type)
	{
		type = type(type);
		
		if (type.equals("int"))
			return "Integer.MIN_VALUE";
		if (type.equals("double"))
			return "Double.NaN";
		if (type.equals("float"))
			return "Float.NaN";
		if (type.equals("long"))
			return "Long.MIN_VALUE";
		if (type.equals("short"))
			return "Short.MIN_VALUE";
		if (type.equals("boolean"))
			return "false";
		if (type.equals("byte"))
			return "Byte.MIN_VALUE";
		if (type.equals("char"))
			return "(char)-1";

		return "null";
	}

	/** Get the default value of a specific type as an Object
	 * @param type Type
	 * @return Default value
	 */
	public static Object defaultVal2 (String type)
	{
		type = type(type);
		
		if (type.equals("int"))
			return new Integer(Integer.MIN_VALUE);
		if (type.equals("double"))
			return new Double(Double.NaN);
		if (type.equals("float"))
			return new Float(Float.NaN);
		if (type.equals("long"))
			return new Long(Long.MIN_VALUE);
		if (type.equals("short"))
			return new Short(Short.MIN_VALUE);
		if (type.equals("boolean"))
			return new Boolean(false);
		if (type.equals("byte"))
			return new Byte(Byte.MIN_VALUE);
		if (type.equals("char"))
			return new Character((char)-1);

		return null;
	}

	/** Type conversion helper function
	 * @param type Type
	 * @return A cleaned-up type
	 */
	public static String convert (String type)
	{
		type = type(type);
		if (type.equals("Integer"))
			type = "Int";
		if (type.equals("Character"))
			type = "Char";
		if ((type.length()>2)&&(type.substring
				(type.length()-2, type.length()).equals("[]")))
			return "Array";
		
		char[] foo = type.toCharArray();
		if ((foo[0]-32)>64)
			foo[0] -= 32;
		return new String(foo);
	}

	/** Retrieves the first matching token from a string
	 * @param str String to look in
	 * @param tok Token to look for
	 * @return First matching token
	 */
	public static String firstToken (String str, String tok)
	{
		StringTokenizer t = new StringTokenizer(str, tok);
		return t.nextToken();
	}

	/** Retrieves the last matching token from a string
	 * @param str String to look in
	 * @param tok Token to look for
	 * @return Last matching token
	 */
	public static String lastToken (String str, String tok)
	{
		String ret = null;
		StringTokenizer t = new StringTokenizer(str, tok);
		while (t.hasMoreTokens())
			ret = t.nextToken();
		return ret;
	}

	/** Checks if a certain token is present in a string
	 * @param str String to look in
	 * @param tok Token to look for
	 * @return True if the token is present, false otherwise
	 */
	public static boolean hasToken (String str, String tok)
	{
		StringTokenizer t = new StringTokenizer(str, tok);
		if (t.countTokens()<2)
			return false;
		return true;
	}

	// TODO: document
	public static String java2cpp (Class klasse)
	{
		if (klasse.equals(Integer.TYPE) || klasse.equals(Integer.class))
			return "int4";
		return "null";
	}

	/** Convert a Type object to a string
	 * @param t Type object
	 * @return String representation of the Type object
	 */
	public static String printType (Type t)
	{
		String ret = handleArray(lastToken(t.toString(), " "));

		// TODO: Yes, this looks better, but might be problematic
		// TODO: Could be hacked to only remove 'java.lang.'
		/*if (hasToken(ret, "<"))
		{
			StringBuffer buf = new StringBuffer();
			buf.append(lastToken(firstToken(ret, "<"), ".")+"<");
			String tmp = lastToken(ret, "<");
			buf.append(lastToken(tmp.substring(0, tmp.length()-1), ".")+">");
			ret = buf.toString();
		}
		else
			ret = lastToken(ret, ".");*/
			
		return ret;
	}

	/* Array conversion functions */
	public static boolean[] booleanArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		boolean[] ret = new boolean[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Boolean)tmp[i]).booleanValue();
		return ret;
	}

	public static byte[] byteArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		byte[] ret = new byte[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Byte)tmp[i]).byteValue();
		return ret;
	}

	public static char[] charArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		char[] ret = new char[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Character)tmp[i]).charValue();
		return ret;
	}
	
	public static double[] doubleArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		double[] ret = new double[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Double)tmp[i]).doubleValue();
		return ret;
	}

	public static float[] floatArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		float[] ret = new float[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Float)tmp[i]).floatValue();
		return ret;
	}

	public static int[] intArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		int[] ret = new int[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Integer)tmp[i]).intValue();
		return ret;
	}

	public static long[] longArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		long[] ret = new long[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Long)tmp[i]).longValue();
		return ret;
	}

	public static short[] shortArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		short[] ret = new short[len];
		for (int i=0;i<len;i++)
			ret[i] = ((Short)tmp[i]).shortValue();
		return ret;
	}

	public static String[] StringArray (Object data)
	{
		Object[] tmp = (Object[])data;
		int len = tmp.length;
		String[] ret = new String[len];
		for (int i=0;i<len;i++)
			ret[i] = (String)tmp[i];
		return ret;
	}

	public static Object[] ObjectArray (Object data)
	{
		return (Object[])data;
	}
}
