package ReflWrap;

import java.util.*;

/** Data structure for keeping track of template types */
public class TypeTree 
{
	/** Type */
	private Class klasse;
	/** Type parameters */
	private LinkedList<TypeTree> targs = new LinkedList<TypeTree>();

	public Class klasse ()
	{
		return klasse;
	}

	public LinkedList<TypeTree> targs ()
	{
		return targs;
	}

	/** String.substring() wrapper 
	 * @param str String
	 * @param start Start of substring
	 * @param end End of substring or -1 for whole string
	 * @return Substring
	 */
	private String substr (String str, int start, int end)
	{	
		if (end == -1)
			return str;
		return str.substring(start, end);
	}

	/** Retrieve the type 
	 * @return Type
	 */
	public Class getType ()
	{
		return klasse;
	}

	/** Retrieve the type parameters
	 * @return Type parameters
	 */
	public LinkedList<TypeTree> getTypeParameters ()
	{
		return targs;
	}

	/** Retrieve the type parameters as array
	 * @return Type parameter array
	 */
	public TypeTree[] getTypeArray ()
	{
		TypeTree[] ret = new TypeTree[targs.size()];
		int i = 0;
		for (TypeTree t : targs)
			ret[i++] = t;
		return ret;
	}

	/** Convenience constructor for simple types with a list of type
	 * parameters which have no further type parameters.
	 * @param args Types
	 */
	public TypeTree (Class... args) throws ClassNotFoundException
	{
		this(TypeTree.classify(args));
	}

	public TypeTree (Class klass, TypeTree tree)
	{
		klasse = klass;
		if (tree != null)
			targs.add(tree);
	}

	public TypeTree (Class klass, LinkedList<TypeTree> tree)
	{
		klasse = klass;
		if (tree != null)
			targs = tree;
	}

	/** Helper function
	 * @param args Types
	 * @return Fully-qualified type names in a string
	 */
	public static String classify (Class[] args)
	{
		if (args.length == 1)
			return (args[0] == null) ? null : args[0].getName();
		String arg = args[0].getName()+"<";
		for (int i=1;i<args.length;i++)
			arg += args[i].getName()+",";
		return arg.substring(0, arg.length()-1)+">";
	}
	
	/** Constructor
	 * @param type Type descriptor
	 */
	public TypeTree (String type) throws ClassNotFoundException
	{
		if (type == null)
			return;
		int idx = type.indexOf("<");
		// Needed because Java's handling of fundamentals is braindead
		String klass = substr(type, 0, idx);
		if (klass.equals("boolean"))
			klass = "java.lang.Boolean";
		if (klass.equals("byte"))
			klass = "java.lang.Byte";
		if (klass.equals("char"))
			klass = "java.lang.Character";
		if (klass.equals("double"))
			klass = "java.lang.Double";
		if (klass.equals("float"))
			klass = "java.lang.Float";
		if (klass.equals("int"))
			klass = "java.lang.Integer";
		if (klass.equals("long"))
			klass = "java.lang.Long";
		if (klass.equals("short"))
			klass = "java.lang.Short";
		klasse = Class.forName(klass);
		
		if (idx == -1)
			return;
		
		String rest = type.substring(idx+1, type.lastIndexOf(">"));
		
		int end = 0;
		while (end != -1)
		{
			end = rest.indexOf(",");
			String cur = substr(rest, 0, end);
			
			if (cur.indexOf("<") != -1)
			{
				end = rest.indexOf(">");
				cur = substr(rest, 0, end++)+">";
			}
			
			TypeTree tree = new TypeTree(cur);
			targs.add(tree);
			if (end >= rest.length())
				break;
			rest = rest.substring(end+1, rest.length());
		}
	}

	/** Retrieve a string represenatation for this object
	 * @return String
	 */
	public String toString ()
	{
		if (targs.size() == 0)
			return klasse.getName();
		String ret = klasse.getName()+"<";
		for (TypeTree t : targs)
			ret += t+",";
		return ret.substring(0, ret.length()-1)+">";
	}

	/** Compare a type descriptor to this TypeTree
	 * @param type Type descriptor
	 * @return True if equal, false if not
	 */
	public boolean equals (String type)
	{
		return toString().equals(type);
	}

	/** Test code */
	public static void main (String[] args)
	{
		try
		{
			String type = "java.util.LinkedList<java.lang.Byte,java.lang.Boolean,java.lang.Byte<java.lang.Class,java.util.HashMap>,java.lang.Double>";
			TypeTree tree = new TypeTree(type); 
			assert(tree.equals(type));

			type = "java.lang.Class";
			tree = new TypeTree(type);
			assert(tree.equals(type));

			type = "java.lang.Class<java.lang.Byte,java.lang.Integer,java.lang.Boolean>";
			tree = new TypeTree(Class.class, Byte.class, Integer.class, 
				Boolean.class);
			assert(tree.equals(type));

			tree = new TypeTree(Byte.class, Integer.class, Boolean.class);
			tree = new TypeTree(Class.class, tree);
			assert(tree.equals(type));
		}
		catch (ClassNotFoundException e)
		{
			e.printStackTrace();
		}
	}
}
